"""Implement a paginated menu.

The first page starts with a heading and its sub-heading.
Each item may have sub items, in that case activating this item will open a new menu
with its sub items.
Each item can optionally be styled differently.
"""

from __future__ import annotations

import math
import pathlib
import threading
import uuid
import warnings
from typing import TYPE_CHECKING, Callable, Sequence, cast, overload

from headless_kivy import HeadlessWidget
from kivy.clock import mainthread
from kivy.lang.builder import Builder
from kivy.properties import (
    AliasProperty,
    BooleanProperty,
    ListProperty,
    NumericProperty,
)
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.screenmanager import Screen, ScreenManager, TransitionBase

from ubo_gui.logger import logger
from ubo_gui.menu._transitions import TransitionsMixin
from ubo_gui.page import PageWidget

from .constants import PAGE_SIZE
from .stack_item import (
    VISUAL_SNAPSHOT_WIDTH,
    StackApplicationItem,
    StackItem,
    StackMenuItem,
    StackMenuItemSelection,
)
from .types import (
    ActionItem,
    ApplicationItem,
    HeadedMenu,
    HeadlessMenu,
    Item,
    Menu,
    SubMenuItem,
    process_subscribable_value,
)
from .widgets.menu_page_widget import MenuPageWidget

if TYPE_CHECKING:
    from kivy.uix.widget import Widget

    from ubo_gui.animated_slider import AnimatedSlider


class MenuWidget(BoxLayout, TransitionsMixin):
    """Paginated menu."""

    menu_subscriptions: set[Callable[[], None]]
    menu_subscriptions_lock: threading.Lock
    screen_subscriptions: set[Callable[[], None]]
    screen_subscriptions_lock: threading.Lock
    stack_lock: threading.Lock

    _current_menu_items: Sequence[Item]
    _current_screen: Screen | None = None
    _title: str | None = None
    screen_manager: ScreenManager
    slider: AnimatedSlider

    def __init__(self: MenuWidget, **kwargs: object) -> None:
        """Initialize a `MenuWidget`."""
        self._current_menu_items = []
        self.menu_subscriptions = set()
        self.menu_subscriptions_lock = threading.Lock()
        self.screen_subscriptions = set()
        self.screen_subscriptions_lock = threading.Lock()
        self.stack_lock = threading.Lock()
        super().__init__(**kwargs)
        self.bind(stack=self._render)

    def __del__(self: MenuWidget) -> None:
        """Clear all subscriptions."""
        self._clear_menu_subscriptions()
        self._clear_screen_subscriptions()
        for item in self.stack:
            item.clear_subscriptions()

    def set_root_menu(self: MenuWidget, root_menu: Menu) -> None:
        """Set the root menu."""
        with self.stack_lock:
            if not self.stack:
                self._push(root_menu, transition=self._no_transition)
            else:
                self._replace_menu(self.root, root_menu)

    def get_depth(self: MenuWidget) -> int:
        """Return depth of the current screen."""
        return len(self.stack)

    def set_title(self: MenuWidget, title: str) -> bool:
        """Set the title of the currently active menu."""
        self._title = title
        return True

    def get_title(self: MenuWidget) -> str | None:
        """Return the title of the currently active menu."""
        if self.current_application:
            return getattr(self.current_application, 'title', None)
        if self.current_menu:
            return self._title
        return None

    @mainthread
    def go_down(self: MenuWidget) -> None:
        """Go to the next page.

        If it is already the last page, rotate to the first page.
        """
        if self.current_application:
            self.current_application.go_down()
            return

        if self.pages == 1:
            return

        if self.current_menu:
            menu_page = cast(PageWidget, self.current_screen)

            menu_page.clone.page_index = self.page_index = (
                self.page_index + 1
            ) % self.pages
            menu_page.clone.items = self._menu_items(self.current_menu)

            self._switch_to(
                menu_page.clone,
                transition=self._slide_transition,
                direction='up',
            )
            self.current_screen = menu_page.clone

    @mainthread
    def go_up(self: MenuWidget) -> None:
        """Go to the previous page.

        If it is already the first page, rotate to the last page.
        """
        if self.current_application:
            self.current_application.go_up()
            return

        if self.pages == 1:
            return

        if self.current_menu:
            menu_page = cast(PageWidget, self.current_screen)

            menu_page.clone.page_index = self.page_index = (
                self.page_index - 1
            ) % self.pages
            menu_page.clone.items = self._menu_items(self.current_menu)

            self._switch_to(
                menu_page.clone,
                transition=self._slide_transition,
                direction='down',
            )
            self.current_screen = menu_page.clone

    def open_menu(
        self: MenuWidget,
        menu: Menu | Callable[[], Menu],
        *,
        key: str = '',
    ) -> None:
        """Open a menu."""
        parent = self.top
        stack_item: StackMenuItem | None = None
        subscription: Callable[[], None] | None = None

        def handle_menu_change(menu: Menu) -> None:
            nonlocal stack_item, subscription
            logger.debug(
                'Handle `sub_menu` change...',
                extra={
                    'new_sub_menu': menu,
                    'old_sub_menu': stack_item.menu if stack_item else None,
                    'subscription_level': 'parent',
                },
            )
            with self.stack_lock:
                if stack_item:
                    stack_item = self._replace_menu(stack_item, menu)
                elif menu:
                    stack_item = self._push(
                        menu,
                        parent=parent,
                        key=key,
                        transition=self._slide_transition,
                        direction='left',
                    )
                    if subscription:
                        stack_item.subscriptions.add(subscription)

        subscription = process_subscribable_value(
            menu,
            handle_menu_change,
        )
        if stack_item:
            stack_item.subscriptions.add(subscription)

    def select_action_item(self: MenuWidget, item: ActionItem) -> None:
        """Select an action item."""
        result = item.action()
        if not result:
            return
        if isinstance(result, type):
            if issubclass(result, PageWidget):
                self.open_application(result())
            else:
                msg = f'Unsupported returned value by `ActionItem`: {result}'
                raise TypeError(msg)
        elif isinstance(result, PageWidget):
            self.open_application(result)
        elif isinstance(result, Menu) or callable(result):
            self.open_menu(result)
        else:
            msg = f'Unsupported returned value by `ActionItem`: {result}'
            raise TypeError(msg)

    def select_application_item(self: MenuWidget, item: ApplicationItem) -> None:
        """Select an application item."""
        application_instance: PageWidget | None = None

        def handle_application_change(application: type[PageWidget]) -> None:
            nonlocal application_instance
            logger.debug(
                'Handle `application` change...',
                extra={
                    'new_application_class': application,
                    'old_application_class': type(application_instance),
                    'old_application': application_instance,
                    'subscription_level': 'parent',
                },
            )
            if application_instance:
                self.close_application(application_instance)
            application_instance = application()
            self.open_application(application_instance)

        self.top.subscriptions.add(
            process_subscribable_value(
                item.application,
                handle_application_change,
            ),
        )

    def select_submenu_item(self: MenuWidget, item: SubMenuItem) -> None:
        """Select a submenu item."""
        if item.key:
            self.open_menu(item.sub_menu, key=item.key)
        else:
            self.open_menu(item.sub_menu)

    def select_item(self: MenuWidget, item: Item) -> None:
        """Select an item.

        Parameters
        ----------
        item: `Item`
            The item to select

        """
        if isinstance(item, ActionItem):
            self.select_action_item(item)
        if isinstance(item, ApplicationItem):
            self.select_application_item(item)
        if isinstance(item, SubMenuItem):
            self.select_submenu_item(item)

    def select(self: MenuWidget, index: int) -> None:
        """Select one of the items currently visible on the screen based on its index.

        Parameters
        ----------
        index: `int`
            An integer number, can only take values greater than or equal to zero and
            less than `PAGE_SIZE`

        """
        if not self.screen_manager.current_screen:
            warnings.warn('`current_screen` is `None`', RuntimeWarning, stacklevel=1)
            return
        if self._is_preparation_in_progress:
            return
        current_page = cast(PageWidget, self.current_screen)
        item = current_page.get_item(index)
        if item:
            self.select_item(item)

    def go_back(self: MenuWidget) -> None:
        """Go back to the previous menu."""
        if self.current_application:
            if not self.current_application.go_back():
                self.close_application(self.current_application)
        elif self.current_menu:
            with self.stack_lock:
                self._pop()

    def go_home(self: MenuWidget) -> None:
        """Go back to the root menu."""
        with self.stack_lock:
            for item in self.stack[1:]:
                item.clear_subscriptions()
                if isinstance(item, StackApplicationItem):
                    item.application.dispatch('on_close')
            self.root.selection = None
            self.stack = self.stack[:1]
            self._switch_to(
                self.current_screen,
                transition=self._rise_in_transition,
            )

    def _menu_items(
        self: MenuWidget,
        menu: Menu,
    ) -> Sequence[Item | None]:
        """Render a normal menu."""
        if self.page_index >= self.pages:
            self.page_index = self.pages - 1
        offset = -(PAGE_SIZE - 1) if isinstance(menu, HeadedMenu) else 0
        items: list[Item | None] = list(
            self.current_menu_items[
                max(self.page_index * PAGE_SIZE + offset, 0) : self.page_index
                * PAGE_SIZE
                + PAGE_SIZE
                + offset
            ],
        )
        if self.render_surroundings:
            previous_item = (
                None
                if self.page_index == 0
                else (
                    padding_item := self.current_menu_items[
                        self.page_index * PAGE_SIZE + offset - 1
                    ]
                )
                and Item(
                    label=padding_item.label,
                    icon=padding_item.icon,
                    background_color=padding_item.background_color,
                    is_short=padding_item.is_short,
                    opacity=0.6,
                )
            )
            next_item = (
                None
                if self.page_index == self.pages - 1
                else (
                    padding_item := self.current_menu_items[
                        self.page_index * PAGE_SIZE + PAGE_SIZE + offset
                    ]
                )
                and Item(
                    label=padding_item.label,
                    icon=padding_item.icon,
                    background_color=padding_item.background_color,
                    is_short=padding_item.is_short,
                    opacity=0.6,
                )
            )
            items = [previous_item, *items, next_item]
        return items

    def _render_menu(self: MenuWidget, menu: Menu) -> MenuPageWidget:
        """Render the items of the current menu."""
        if self.page_index >= self.pages:
            self.page_index = self.pages - 1
        items = self._menu_items(menu)

        list_widget = MenuPageWidget(
            items,
            page_index=self.page_index,
            name=f'Page {self.get_depth()} {self.page_index}',
            count=PAGE_SIZE,
            render_surroundings=self.render_surroundings,
            padding_bottom=self.padding_bottom,
            padding_top=self.padding_top,
        )

        if isinstance(self.current_menu, HeadedMenu):

            def handle_heading_change(heading: str) -> None:
                logger.debug(
                    'Handle `heading` change...',
                    extra={
                        'new_heading': heading,
                        'old_heading': list_widget.heading,
                        'subscription_level': 'widget',
                    },
                )
                list_widget.heading = heading

            self.menu_subscriptions.add(
                process_subscribable_value(
                    self.current_menu.heading,
                    handle_heading_change,
                ),
            )

            def handle_sub_heading_change(sub_heading: str) -> None:
                logger.debug(
                    'Handle `sub_heading` change...',
                    extra={
                        'new_sub_heading': sub_heading,
                        'old_sub_heading': list_widget.sub_heading,
                        'subscription_level': 'widget',
                    },
                )
                list_widget.sub_heading = sub_heading

            self.menu_subscriptions.add(
                process_subscribable_value(
                    self.current_menu.sub_heading,
                    handle_sub_heading_change,
                ),
            )

        return list_widget

    def _render(self: MenuWidget, *_: object) -> None:
        """Return the current screen page."""
        self._clear_screen_subscriptions()

        if not self.stack:
            return

        title = None
        if isinstance(self.top, StackApplicationItem):
            self.current_screen = self.top.application
            title = self.top.application.title
        if isinstance(self.top, StackMenuItem):
            menu = self.top.menu
            last_items = None
            menu_page: MenuPageWidget | None = None
            placeholder = None

            def handle_items_change(items: Sequence[Item]) -> None:
                nonlocal last_items, menu_page
                logger.debug(
                    'Handle `items` change...',
                    extra={
                        'new_items': items,
                        'old_items': last_items,
                        'subscription_level': 'screen',
                    },
                )
                self.current_menu_items = items
                if menu_page is None:
                    self._clear_menu_subscriptions()
                    menu_page = self._render_menu(menu)
                    # The clone here is solely needed for the visual transitions between
                    # menu pages, when `page_index` increases or decreases, the slide
                    # down/up transition needs a source and a target. So a clone of the
                    # original page is needed. If `ScreenManager` supported transition
                    # from a screen to itself we probably wouldn't need this.
                    menu_page.clone = self._render_menu(menu)
                    menu_page.clone.clone = menu_page
                    self.current_screen = menu_page
                else:
                    if self.page_index >= self.pages:
                        menu_page.page_index = self.page_index = self.pages - 1
                        menu_page.clone.page_index = self.page_index = self.pages - 1
                    menu_page.items = self._menu_items(menu)
                    menu_page.clone.items = self._menu_items(menu)
                menu_page.placeholder = placeholder
                menu_page.clone.placeholder = placeholder
                last_items = items

            self.screen_subscriptions.add(
                process_subscribable_value(menu.items, handle_items_change),
            )

            def handle_placeholder_change(new_placeholder: str | None) -> None:
                nonlocal placeholder
                logger.debug(
                    'Handle `placeholder` change...',
                    extra={
                        'new_placeholder': new_placeholder,
                        'old_placeholder': placeholder,
                        'subscription_level': 'widget',
                    },
                )
                placeholder = new_placeholder
                if menu_page:
                    menu_page.placeholder = placeholder

            self.menu_subscriptions.add(
                process_subscribable_value(
                    menu.placeholder,
                    handle_placeholder_change,
                ),
            )

            title = menu.title

        def handle_title_change(title: str | None) -> None:
            logger.debug(
                'Handle `title` change...',
                extra={
                    'new_title': title,
                    'old_title': self.title,
                    'subscription_level': 'screen',
                },
            )
            self.title = title

        self.screen_subscriptions.add(
            process_subscribable_value(title, handle_title_change),
        )

    def get_current_screen(self: MenuWidget) -> Screen | None:
        """Return current screen."""
        return self._current_screen

    def _get_current_screen(self: MenuWidget) -> Screen | None:
        """Workaround for `AliasProperty` not working with overridden getters."""
        return self.get_current_screen()

    def set_current_screen(self: MenuWidget, screen: Screen) -> bool:
        """Set the current screen page."""
        self._current_screen = screen
        return True

    def open_application(self: MenuWidget, application: PageWidget) -> None:
        """Open an application."""
        with self.stack_lock:
            headless_widget = HeadlessWidget.get_instance(self)
            if headless_widget:
                headless_widget.activate_high_fps_mode()
            application.name = uuid.uuid4().hex
            application.padding_bottom = self.padding_bottom
            application.padding_top = self.padding_top
            self._push(
                application,
                transition=self._swap_transition,
                duration=0.2,
                direction='left',
            )

    def close_application(self: MenuWidget, application: PageWidget) -> None:
        """Close an application after its `on_close` event is fired."""
        # Remove `application` and all applications in the stack with their `root` being
        # `application` from stack and clear their bindings and subscriptions.
        # If any of these applications are the top of the stack, remove it with `pop` to
        # ensure the animation is played.
        with self.stack_lock:
            if any(
                isinstance(item.root, StackApplicationItem)
                and item.root.application is application
                for item in self.stack
            ):
                to_be_removed = [
                    cast(StackApplicationItem, item)
                    for item in self.stack
                    if isinstance(item.root, StackApplicationItem)
                    and item.root.application is application
                    and item is not self.top
                ]

                for item in to_be_removed:
                    item.clear_subscriptions()
                    item.application.dispatch('on_close')

                self.stack = [item for item in self.stack if item not in to_be_removed]

                if (
                    isinstance(self.top.root, StackApplicationItem)
                    and self.top.root.application is application
                ):
                    self._pop()

    @property
    def root(self: MenuWidget) -> StackMenuItem:
        """Return the root item."""
        if isinstance(self.stack[0], StackMenuItem):
            return self.stack[0]
        msg = 'root is not a `StackMenuItem`'
        raise ValueError(msg)

    @property
    def top(self: MenuWidget) -> StackItem:
        """Return the top item of the stack."""
        if not self.stack:
            msg = 'stack is empty'
            raise IndexError(msg)
        return self.stack[-1]

    @property
    def _visual_snapshot(self: MenuWidget) -> list[str]:
        start = [
            '╭',
            '│',
            '│',
            '│',
            '╰',
        ]
        end = [
            '╮',
            '│',
            '│',
            '│',
            '╯',
        ]

        cross_repeats = VISUAL_SNAPSHOT_WIDTH // 3
        cross = [
            '───' * cross_repeats,
            '╲ ╱' * cross_repeats,  # noqa: RUF001
            ' ╳ ' * cross_repeats,  # noqa: RUF001
            '╱ ╲' * cross_repeats,  # noqa: RUF001
            '───' * cross_repeats,
        ]

        def append(item: list[str]) -> None:
            for i in range(5):
                output[-(5 - i)] += item[i] if i < len(item) else ' ' * len(item[0])

        output = []
        for start_item in range(0, len(self.stack), 5):
            output.extend([''] * 5)
            for item in self.stack[start_item : min(start_item + 5, len(self.stack))]:
                append(start)
                append(item.visual_snapshot)
            append(end)
            append(
                [
                    f' {type(item).__name__} {item.title[:VISUAL_SNAPSHOT_WIDTH]} '
                    for item in self.stack[
                        start_item : min(start_item + 5, len(self.stack))
                    ]
                ],
            )

            output.extend([''] * 5)
            for item in self.stack[start_item : min(start_item + 5, len(self.stack))]:
                append(start)
                append(item.parent.visual_snapshot if item.parent else cross)
            append(end)

            output.extend([''] * 5)
            for item in self.stack[start_item : min(start_item + 5, len(self.stack))]:
                append(start)
                append(
                    item.selection.item.visual_snapshot
                    if isinstance(item, StackMenuItem) and item.selection
                    else cross,
                )
            append(end)
        if len(self.stack) % 5 != 0:
            for _ in range(5 - (len(self.stack) % 5)):
                append([' ' * (VISUAL_SNAPSHOT_WIDTH + 1)] * 5)

        return output

    def _replace_menu(
        self: MenuWidget,
        stack_item: StackMenuItem,
        menu: Menu,
        *,
        parent: StackItem | None = None,
    ) -> StackMenuItem:
        """Replace the current menu or application."""
        if stack_item not in self.stack:
            msg = '`stack_item` not found in stack'
            raise ValueError(msg) from None
        if stack_item.selection:
            items = [
                item
                for item in (menu.items() if callable(menu.items) else menu.items)
                if isinstance(item, SubMenuItem)
                and item.key == stack_item.selection.key
            ]
            if len(items) == 1:
                selection = items[0]
            elif len(items) > 1:
                msg = f'Found more than one item with key: {stack_item.selection.key}'
                raise ValueError(msg)
            else:
                selection = None
        else:
            selection = None
        index = self.stack.index(stack_item)
        new_item = self.stack[index] = StackMenuItem(
            menu=menu,
            page_index=stack_item.page_index,
            parent=parent or stack_item.parent,
            subscriptions=stack_item.subscriptions,
        )
        new_item.selection = (
            None
            if selection is None
            or stack_item.selection is None
            or not isinstance(selection.sub_menu, Menu)
            else StackMenuItemSelection(
                key=stack_item.selection.key,
                item=self._replace_menu(
                    stack_item.selection.item,
                    selection.sub_menu,
                    parent=new_item,
                ),
            )
        )

        if new_item is self.top:
            self._switch_to(self.current_screen, transition=self._no_transition)
        return new_item

    @overload
    def _push(
        self: MenuWidget,
        item: Menu,
        /,
        *,
        transition: TransitionBase | None,
        duration: float | None = None,
        direction: str | None = None,
        parent: StackItem | None = None,
        key: str | None = None,
    ) -> StackMenuItem: ...
    @overload
    def _push(  # pyright: ignore[reportOverlappingOverload]
        self: MenuWidget,
        item: PageWidget,
        /,
        *,
        transition: TransitionBase | None,
        duration: float | None = None,
        direction: str | None = None,
        parent: StackItem | None = None,
        key: str | None = None,
    ) -> StackApplicationItem: ...
    def _push(  # noqa: PLR0913
        self: MenuWidget,
        item: Menu | PageWidget,
        /,
        *,
        transition: TransitionBase | None,
        duration: float | None = None,
        direction: str | None = None,
        parent: StackItem | None = None,
        key: str | None = None,
    ) -> StackItem:
        """Go one level deeper in the menu stack."""
        if isinstance(item, Menu):
            new_top = StackMenuItem(menu=item, page_index=0, parent=parent)
        elif isinstance(item, PageWidget):
            new_top = StackApplicationItem(application=item, parent=parent)
        else:
            msg = f'Unsupported type: {type(item)}'
            raise TypeError(msg)

        if isinstance(parent, StackMenuItem) and isinstance(new_top, StackMenuItem):
            parent.selection = StackMenuItemSelection(key=key or '', item=new_top)

        self.stack = [*self.stack, new_top]

        self._switch_to(
            self.current_screen,
            transition=transition,
            duration=duration,
            direction=direction,
        )

        return new_top

    def _pop(
        self: MenuWidget,
        /,
        *,
        transition: TransitionBase | None = None,
        duration: float | None = None,
        direction: str | None = 'right',
        keep_subscriptions: bool = False,
    ) -> None:
        """Come up one level from of the menu stack."""
        if self.depth == 1:
            return
        popping_item = self.top
        if not keep_subscriptions:
            popping_item.clear_subscriptions()
        if isinstance(popping_item, StackApplicationItem):
            popping_item.application.dispatch('on_close')
        if isinstance(popping_item.parent, StackMenuItem):
            popping_item.parent.selection = None

        *self.stack, _ = self.stack

        target = self.top

        transition_ = self._slide_transition
        if isinstance(target, StackApplicationItem) or self.current_application:
            transition_ = self._swap_transition
        self._switch_to(
            self.current_screen,
            transition=transition or transition_,
            duration=duration,
            direction=direction,
        )

    def get_is_scrollbar_visible(self: MenuWidget) -> bool:
        """Return whether scroll-bar is needed or not."""
        return not self.current_application and self.pages > 1

    def on_kv_post(self: MenuWidget, base_widget: Widget) -> None:
        """Run after the widget is fully constructed."""
        _ = base_widget
        self.screen_manager = cast(ScreenManager, self.ids.screen_manager)
        self.slider = self.ids.slider

    def _clear_menu_subscriptions(self: MenuWidget) -> None:
        """Clear widget subscriptions."""
        with self.menu_subscriptions_lock:
            subscriptions = self.menu_subscriptions.copy()
            self.menu_subscriptions.clear()
            for unsubscribe in subscriptions:
                unsubscribe()

    def _clear_screen_subscriptions(self: MenuWidget) -> None:
        """Clear screen subscriptions."""
        # lock the mutex to do it atomic
        with self.screen_subscriptions_lock:
            subscriptions = self.screen_subscriptions.copy()
            self.screen_subscriptions.clear()
            for unsubscribe in subscriptions:
                unsubscribe()

    def get_current_application(self: MenuWidget) -> PageWidget | None:
        """Return the current application."""
        if self.stack and isinstance(self.top, StackApplicationItem):
            return self.top.application
        return None

    def get_current_menu(self: MenuWidget) -> Menu | None:
        """Return the current menu."""
        if self.stack and isinstance(self.top, StackMenuItem):
            return self.top.menu
        return None

    def get_page_index(self: MenuWidget) -> int:
        """Return the current page index."""
        if self.stack and isinstance(self.top, StackMenuItem):
            return self.top.page_index
        return 0

    def set_page_index(self: MenuWidget, page_index: int) -> bool:
        """Set the current page index."""
        if self.stack and isinstance(self.top, StackMenuItem):
            if self.top.page_index != page_index:
                self.top.page_index = page_index
                return True
            return False
        return True

    def get_pages(self: MenuWidget) -> int:
        """Return the number of pages of the currently active menu."""
        if isinstance(self.current_menu, HeadedMenu):
            return max(math.ceil((len(self.current_menu_items) + 2) / 3), 1)
        if isinstance(self.current_menu, HeadlessMenu):
            return max(math.ceil(len(self.current_menu_items) / 3), 1)
        return 0

    def get_current_menu_items(self: MenuWidget) -> Sequence[Item] | None:
        """Return current menu items."""
        return self._current_menu_items

    def set_current_menu_items(self: MenuWidget, items: Sequence[Item]) -> bool:
        """Set current menu items."""
        self._current_menu_items = items
        self.slider.value = self.get_pages() - 1 - self.page_index
        return True

    stack: list[StackItem] = ListProperty()
    title = AliasProperty(
        getter=get_title,
        setter=set_title,
        bind=['stack'],
    )
    depth: int = AliasProperty(getter=get_depth, bind=['stack'], cache=True)
    pages: int = AliasProperty(
        getter=get_pages,
        bind=['current_menu_items', 'current_menu'],
        cache=True,
    )
    page_index = AliasProperty(
        getter=get_page_index,
        setter=set_page_index,
        bind=['current_menu_items'],
    )
    current_menu_items: Sequence[Item] = AliasProperty(
        getter=get_current_menu_items,
        setter=set_current_menu_items,
    )
    current_application: PageWidget | None = AliasProperty(
        getter=get_current_application,
        bind=['stack'],
        cache=True,
    )
    current_menu: Menu | None = AliasProperty(
        getter=get_current_menu,
        bind=['stack'],
        cache=True,
    )
    current_menu_type: type[Menu] | None = AliasProperty(
        getter=lambda self: type(self.current_menu) if self.current_menu else None,
        bind=['current_menu'],
        cache=True,
    )
    current_screen: Screen | None = AliasProperty(
        getter=_get_current_screen,
        setter=set_current_screen,
    )
    is_scrollbar_visible = AliasProperty(
        getter=get_is_scrollbar_visible,
        bind=['pages'],
        cache=True,
    )
    render_surroundings = BooleanProperty(
        defaultvalue=False,
        cache=True,
    )
    padding_bottom = NumericProperty(defaultvalue=0)
    padding_top = NumericProperty(defaultvalue=0)

    def __repr__(self: MenuWidget) -> str:
        """Return a string representation of the widget."""
        return '\n'.join(self._visual_snapshot)


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('menu.kv').resolve().as_posix(),
)



"""QRCodeWidget class module."""

from __future__ import annotations

import io

import qrcode
from kivy.core.image import Image as CoreImage
from kivy.properties import OptionProperty, StringProperty
from kivy.uix.image import Image


class QRCodeWidget(Image):
    """A widget to display a QR code."""

    content: str = StringProperty()
    fit_mode = OptionProperty(
        'contain',
        options=['scale-down', 'fill', 'contain', 'cover'],
    )

    def on_content(self: QRCodeWidget, _: QRCodeWidget, value: str) -> None:
        """Handle the `content` property change."""
        img = qrcode.make(value, version=1, border=1)

        data = io.BytesIO()
        img.save(data)
        data.seek(0)

        core_image = CoreImage(data, ext='png')
        texture = core_image.texture

        self.texture = texture



"""Class definition of main datatypes use in menus."""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Protocol,
    Sequence,
    TypeAlias,
    TypeGuard,
    cast,
    overload,
)

from immutable import Immutable
from typing_extensions import TypeVar

from ubo_gui.constants import PRIMARY_COLOR

if TYPE_CHECKING:
    from kivy.graphics.context_instructions import Color

    from ubo_gui.page import PageWidget


class BaseMenu(Immutable):
    """A class used to represent a menu.

    Attributes
    ----------
    title: `str`
        Rendered on top of the menu in all pages. Optionally it can be a callable
    returning the title.

    items: `list` of `Item`
        List of the items of the menu. Optionally it can be a callable returning the
    list of items

    """

    title: str | Callable[[], str]
    items: Sequence[Item] | Callable[[], Sequence[Item]]
    placeholder: str | None | Callable[[], str | None] = None


class HeadedMenu(BaseMenu):
    """A class used to represent a headed menu.

    Attributes
    ----------
    heading: `str`
        Rendered in the first page of the menu, stating the purpose of the menu and its
    items. Optionally it can be a callable returning the heading.

    sub_heading: `str`
        Rendered beneath the heading in the first page of the menu with a smaller font.
    Optionally it can be a callable returning the sub heading.

    """

    heading: str | Callable[[], str]
    sub_heading: str | Callable[[], str]


class HeadlessMenu(BaseMenu):
    """A class used to represent a headless menu."""


Menu: TypeAlias = HeadedMenu | HeadlessMenu


def menu_items(menu: Menu | None) -> Sequence[Item]:
    """Return items of the menu.

    in case it's a function, the return value of the function is called.
    """
    if not menu:
        return []
    return menu.items() if callable(menu.items) else menu.items


class Item(Immutable):
    """A class used to represent a menu item.

    Attributes
    ----------
    label: `str`
        The label of the item. Optionally it can be a callable returning the label.

    color: `str` or `tuple` of `float`
        The color by its name or in rgba format as a list of floats, the list should
    contain 4 elements: red, green, blue and alpha, each being a number in the range
    [0..1].
        For example (0.5, 0, 0.5, 0.8) represents a semi transparent purple, or the
    string 'yellow' represents color yellow.
        Optionally it can be a callable returning the color.

    background_color: `str` or `tuple` of `float`
        The background color by color name or in rgba format as a list of floats, the
    list should contain 4 elements red, green, blue and alpha, each being a number in
    the range [0..1].
        For example (0.5, 0, 0.5, 0.8) represents a semi transparent purple, or the
    string 'yellow' represents color yellow.
        Optionally it can be a callable returning the background color.

    icon: `str`
        Name of a Material Symbols icon, check here for the complete list:
    https://fonts.google.com/icons
        Optionally it can be a callable returning the icon.

    is_short: `bool`
        Whether the item should be rendered in short form or not. In short form only the
    icon of the item is rendered and its label is hidden.
        Optionally it can be a callable returning the is_short value.

    opacity: `float`
        Between 0 and 1, sets the transparency channel of the item.

    """

    key: str | None = None
    label: str | Callable[[], str] = ''
    color: Color | Callable[[], Color] = (1, 1, 1, 1)
    background_color: Color | Callable[[], Color] = PRIMARY_COLOR
    icon: str | None | Callable[[], str | None] = None
    is_short: bool | Callable[[], bool] = False
    opacity: float | None = None
    progress: float | None = None


class ActionItem(Item):
    """A class used to represent an action menu item.

    Attributes
    ----------
    action: `Function`
        If provided, activating this item will call this function.

    """

    action: Callable[
        [],
        Menu | Callable[[], Menu] | type[PageWidget] | PageWidget | None,
    ]


class ApplicationItem(Item):
    """A class used to represent an application menu item.

    Attributes
    ----------
    application: `PageWidget`
        If provided, activating this item will show this widget. Optionally it can be a
    callable returning the widget.

    """

    application: type[PageWidget] | Callable[[], type[PageWidget]]


class SubMenuItem(Item):
    """A class used to represent a sub-menu menu item.

    Attributes
    ----------
    sub_menu: `Menu`
        If provided, activating this item will open another menu, the description
        described in this field. Optionally it can be a callable returning the menu.

    """

    sub_menu: Menu | Callable[[], Menu]


T = TypeVar('T', infer_variance=True)


class Subscribable(Protocol, Generic[T]):
    """A callable that can be subscribed to."""

    def subscribe(
        self: Subscribable,
        callback: Callable[[T], Any],
    ) -> Callable[[], None]:
        """Subscribe to the changes."""
        ...


def is_subscribeable(value: T | Callable[[], T]) -> TypeGuard[Subscribable[T]]:
    """Check if the value is subscribable."""
    return callable(value) and hasattr(value, 'subscribe')


@overload
def process_subscribable_value(
    value: T | Callable[[], T],
    callback: Callable[[T], None],
) -> Callable[[], None]: ...
@overload
def process_subscribable_value(
    value: T | None | Callable[[], T | None],
    callback: Callable[[T | None], None],
) -> Callable[[], None]: ...
def process_subscribable_value(
    value: T | None | Callable[[], T],
    callback: Callable[[T], None],
) -> Callable[[], None]:
    """Return the attribute of the menu or item.

    in case it's a function, the return value of the function is called.
    """
    if is_subscribeable(value):
        return cast(Subscribable[T], value).subscribe(callback)
    processed_value = cast(
        T,
        value() if callable(value) and not isinstance(value, type) else value,
    )
    callback(processed_value)
    return lambda: None



"""Module containing the `PageWidget` class."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Self, Sequence

from kivy.properties import (
    AliasProperty,
    BooleanProperty,
    ListProperty,
    NumericProperty,
    StringProperty,
)
from kivy.uix.screenmanager import Screen

if TYPE_CHECKING:
    from ubo_gui.menu.types import Item


PAGE_MAX_ITEMS = 3


class PageWidget(Screen):
    """A `Screen` that represents a page in the menu."""

    __events__ = ('on_close',)

    items: Sequence[Item | None] = ListProperty([])
    title: str | None = StringProperty(allownone=True, defaultvalue=None)
    count: int = NumericProperty(defaultvalue=PAGE_MAX_ITEMS)
    placeholder: str | None = StringProperty(allownone=True)
    render_surroundings: bool = BooleanProperty(
        defaultvalue=False,
        cache=True,
    )
    padding_top: int = NumericProperty(defaultvalue=0)
    padding_bottom: int = NumericProperty(defaultvalue=0)

    def get_is_empty(self: PageWidget) -> bool:
        """Check if there is no item in items or all of them are `None`."""
        return all(i is None for i in self.items)

    is_empty: bool = AliasProperty(getter=get_is_empty, bind=('items',))

    @property
    def _count(self: PageWidget) -> int:
        return self.count + (2 if self.render_surroundings else 0)

    @property
    def _offset(self: PageWidget) -> int:
        return 1 if self.render_surroundings else 0

    def go_up(self: Self) -> None:
        """Implement this method to provide custom logic for up key."""
        return

    def go_down(self: Self) -> None:
        """Implement this method to provide custom logic for down key."""
        return

    def go_back(self: Self) -> bool | None:
        """Implement this method to provide custom logic for back key.

        Return `true` if back button is handled and the default behavior should be
        avoided
        """
        return

    def __init__(
        self: PageWidget,
        items: Sequence[Item | None] | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        """Initialize a `PageWidget`.

        Parameters
        ----------
        items: `list` of `Item`
            The items to be shown in this page

        args: object
            Stuff that will get directly passed to the `__init__` method of Kivy's

        kwargs: object
            Stuff that will get directly passed to the `__init__` method of Kivy's
        `Screen`.

        """
        self.items = items or []
        super().__init__(*args, **kwargs)
        if items and len(items) > self._count:
            msg = f"""`PageWidget` is initialized with more than `count`={
            self.count} items"""
            raise ValueError(msg)

    def get_item(self: PageWidget, index: int) -> Item | None:
        """Get the page item at the given index."""
        index += self._offset
        if not 0 <= index < len(self.items):
            msg = f"""index must be greater than or equal to 0 and less than {
            len(self.items)}"""
            warnings.warn(msg, ResourceWarning, stacklevel=1)
            return None
        return self.items[index]

    def on_close(self: PageWidget) -> None:
        """Signal when the page is closed."""

    def __repr__(self: PageWidget) -> str:
        """Return a string representation of the `PageWidget`."""
        return (
            f'<{self.__class__.__name__}@PageWidget name="{self.name}" '
            f'title="{self.title}">'
        )



"""Initializes the ubo_gui package and registers various widgets using Kivy Factory.

It also defines the paths for assets and fonts used by the package.
"""

from pathlib import Path

ROOT_PATH = Path(__file__).parent
ASSETS_PATH = ROOT_PATH.joinpath('assets')
FONTS_PATH = ASSETS_PATH.joinpath('fonts')


def setup() -> None:
    """Register various widgets using Kivy Factory and sets up the constants."""
    from kivy.factory import Factory
    from kivy.lang.builder import Builder

    from ubo_gui import constants
    from ubo_gui.menu import constants as menu_constants

    Builder.load_string(
        f"""
    #:set UBO_GUI_PRIMARY_COLOR '{constants.PRIMARY_COLOR}'
    #:set UBO_GUI_SECONDARY_COLOR '{constants.SECONDARY_COLOR}'
    #:set UBO_GUI_SECONDARY_COLOR_LIGHT '{constants.SECONDARY_COLOR_LIGHT}'
    #:set UBO_GUI_TEXT_COLOR '{constants.TEXT_COLOR}'

    #:set UBO_GUI_INFO_COLOR '{constants.INFO_COLOR}'
    #:set UBO_GUI_DANGER_COLOR '{constants.DANGER_COLOR}'
    #:set UBO_GUI_WARNING_COLOR '{constants.WARNING_COLOR}'
    #:set UBO_GUI_SUCCESS_COLOR '{constants.SUCCESS_COLOR}'

    #:set UBO_GUI_PAGE_SIZE {menu_constants.PAGE_SIZE}
    #:set UBO_GUI_SHORT_WIDTH {menu_constants.SHORT_WIDTH}
    #:set UBO_GUI_MENU_ITEM_HEIGHT {menu_constants.MENU_ITEM_HEIGHT}
    #:set UBO_GUI_MENU_ITEM_GAP {menu_constants.MENU_ITEM_GAP}
    """,
    )

    Factory.register('AnimatedSlider', module='ubo_gui.animated_slider')
    Factory.register('GaugeWidget', module='ubo_gui.gauge')
    Factory.register('ItemWidget', module='ubo_gui.menu.widgets.item_widget')
    Factory.register('MenuWidget', module='ubo_gui.menu')
    Factory.register('NotificationWidget', module='ubo_gui.notification')
    Factory.register('ProgressRingWidget', module='ubo_gui.progress_ring')
    Factory.register('PromptWidget', module='ubo_gui.prompt')
    Factory.register('QRCodeWidget', module='ubo_gui.qrcode')
    Factory.register('VolumeWidget', module='ubo_gui.volume')



"""Notification widget."""

from __future__ import annotations

import pathlib

from kivy.lang.builder import Builder
from kivy.metrics import dp
from kivy.properties import ColorProperty, StringProperty

from ubo_gui.page import PageWidget


class NotificationWidget(PageWidget):
    """renders a notification."""

    notification_title: str = StringProperty()
    content: str = StringProperty()
    icon: str = StringProperty()
    color = ColorProperty()

    def go_down(self: NotificationWidget) -> None:
        """Scroll down the notification list."""
        self.ids.slider.animated_value -= dp(50)

    def go_up(self: NotificationWidget) -> None:
        """Scroll up the notification list."""
        self.ids.slider.animated_value += dp(50)


Builder.load_file(
    pathlib.Path(__file__)
    .parent.joinpath('notification_widget.kv')
    .resolve()
    .as_posix(),
)



"""Module for the `AnimatedSlider` class."""

from __future__ import annotations

import pathlib

from kivy.animation import Animation
from kivy.lang.builder import Builder
from kivy.properties import AliasProperty
from kivy.uix.slider import Slider


class AnimatedSlider(Slider):
    """A slider that moves up and down when its value is changed."""

    def set_animated_value(self: AnimatedSlider, value: float) -> bool:
        """Set the value of the `animated_value` property.

        Animates the slider moving to the new value.

        Arguments:
        ---------
        value: The new value that the `animated_value` property is being set to.

        """
        self._animated_value = max(self.min, min(value, self.max))
        Animation(value=self._animated_value, duration=0.2).start(self)
        return True

    def get_animated_value(self: AnimatedSlider) -> float:
        """Return the value of the `animated_value` property.

        Returns
        -------
            The value of the `animated_value` property.

        """
        return self._animated_value

    _animated_value = 0
    animated_value = AliasProperty(
        setter=set_animated_value,
        getter=get_animated_value,
    )

    def __init__(self: AnimatedSlider, **kwargs: object) -> None:
        """Initialize the `AnimatedSlider` class."""
        super().__init__(**kwargs)
        self.value = self.animated_value


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('animated_slider.kv').resolve().as_posix(),
)



"""Module to store constants used in the application."""

PRIMARY_COLOR = '#68B7FF'
SECONDARY_COLOR = '#363F4B'
SECONDARY_COLOR_LIGHT = '#ABA7A7'
TEXT_COLOR = '#FFFFFF'

INFO_COLOR = '#2196F3'
DANGER_COLOR = '#FF3F51'
WARNING_COLOR = '#FFC107'
SUCCESS_COLOR = '#03F7AE'



"""ProgressRing widget."""

from __future__ import annotations

import pathlib

from kivy.lang.builder import Builder
from kivy.metrics import dp
from kivy.properties import ColorProperty, NumericProperty
from kivy.uix.widget import Widget


class ProgressRingWidget(Widget):
    """renders a progress ring."""

    progress: float = NumericProperty(defaultvalue=0, min=0, max=1)
    background_band_width: int = NumericProperty(defaultvalue=dp(1))
    band_width: int = NumericProperty(defaultvalue=dp(4))
    background_color = ColorProperty()
    color = ColorProperty()

    _progress: int = NumericProperty()


Builder.load_file(
    pathlib.Path(__file__)
    .parent.joinpath('progress_ring_widget.kv')
    .resolve()
    .as_posix(),
)



"""Module for the `NormalMenuPageWidget` class."""

from __future__ import annotations

import pathlib
import warnings
from typing import TYPE_CHECKING, Any, Sequence

from kivy.lang.builder import Builder
from kivy.properties import AliasProperty, NumericProperty, StringProperty

from ubo_gui.menu.widgets.item_widget import ItemWidget
from ubo_gui.page import PageWidget

if TYPE_CHECKING:
    from ubo_gui.menu.types import Item


HEADER_SIZE = 2


class MenuPageWidget(PageWidget):
    """renders a page of a `Menu`."""

    def get_has_heading(self: MenuPageWidget) -> bool:
        """Check if the page is rendering a heading."""
        return self.page_index == 0 and bool(self.heading or self.sub_heading)

    heading: str | None = StringProperty(allownone=True, default=None)
    sub_heading: str | None = StringProperty(allownone=True, default=None)
    page_index: int = NumericProperty()
    has_heading: bool = AliasProperty(
        getter=get_has_heading,
        bind=('heading', 'sub_heading', 'page_index'),
    )

    @property
    def head_size(self: MenuPageWidget) -> int:
        """Return the size of the header."""
        return HEADER_SIZE if self.has_heading else 0

    @property
    def _count(self: MenuPageWidget) -> int:
        return super()._count - self.head_size

    def __init__(
        self: MenuPageWidget,
        items: Sequence[Item | None] | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Initialize `MenuPageWidget`.

        Parameters
        ----------
        items: `Sequence`[[`Item`]]
            The item to be shown in this page

        kwargs: Any
            Stuff that will get directly passed to the `__init__` method of Kivy's
        `Screen`.

        """
        self.item_widgets: list[ItemWidget] = []
        self.bind(on_kv_post=self.adjust_item_widgets)
        super().__init__(items, **kwargs)
        self.bind(on_count=self.adjust_item_widgets)
        self.bind(heading=self.adjust_item_widgets)
        self.bind(sub_heading=self.adjust_item_widgets)
        self.bind(page_index=self.adjust_item_widgets)
        self.bind(items=self.render)

    def adjust_item_widgets(self: MenuPageWidget, *args: object) -> None:
        """Initialize the widget."""
        _ = args
        offset = self._offset if self.has_heading else 0
        for _ in range(len(self.item_widgets), self._count - offset):
            self.item_widgets.append(ItemWidget(size_hint=(1, None)))
            self.ids.layout.add_widget(self.item_widgets[-1])
        for _ in range(self._count - offset, len(self.item_widgets)):
            self.ids.layout.remove_widget(self.item_widgets[-1])
            del self.item_widgets[-1]
        self.render()

    def render(self: MenuPageWidget, *_: object) -> None:
        """Render the item widgets."""
        if not self.item_widgets:
            return
        if self.has_heading:
            for i in range(self._offset, self._count):
                self.item_widgets[i - self._offset].item = (
                    self.items[i] if i < len(self.items) else None
                )
        else:
            for i in range(self._count):
                self.item_widgets[i].item = (
                    self.items[i] if i < len(self.items) else None
                )

    def get_item(self: MenuPageWidget, index: int) -> Item | None:
        """Get the item at the given index."""
        if index < self.head_size:
            warnings.warn(
                f'index must be greater than or equal to {self.head_size}',
                ResourceWarning,
                stacklevel=1,
            )
            return None
        try:
            return self.items[index + self._offset - self.head_size]
        except IndexError:
            return None


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('menu_page_widget.kv').resolve().as_posix(),
)



"""A widget that displays a volume bar."""

from __future__ import annotations

import pathlib

from kivy.lang.builder import Builder
from kivy.properties import ColorProperty, NumericProperty
from kivy.uix.boxlayout import BoxLayout

from ubo_gui.constants import PRIMARY_COLOR, SECONDARY_COLOR


class VolumeWidget(BoxLayout):
    """A widget that displays a volume bar."""

    value = NumericProperty(50)
    fill_color = ColorProperty(PRIMARY_COLOR)
    background_color = ColorProperty(SECONDARY_COLOR)


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('volume_widget.kv').resolve().as_posix(),
)



"""Implements the `PromptWidget` class.

A widget that renders a prompt with two options.
"""

from __future__ import annotations

import pathlib
import warnings
from abc import ABC, ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Sequence

from kivy.lang.builder import Builder
from kivy.properties import (
    AliasProperty,
    BooleanProperty,
    ColorProperty,
    StringProperty,
)
from kivy.uix.widget import WidgetMetaclass

from ubo_gui.constants import DANGER_COLOR, SUCCESS_COLOR
from ubo_gui.menu.types import ActionItem
from ubo_gui.page import PageWidget

if TYPE_CHECKING:
    from ubo_gui.menu.types import Item, Menu

PROMPT_OPTIONS = 2


class PromptWidgetMetaClass(ABCMeta, WidgetMetaclass):
    """Metaclass merging `ABC` and `PageWidget` for `PromptWidget` class."""


class PromptWidget(PageWidget, ABC, metaclass=PromptWidgetMetaClass):
    """A widget that renders a prompt."""

    icon: str = StringProperty()
    prompt: str = StringProperty()

    first_option_label: str = StringProperty()
    first_option_icon: str = StringProperty()
    first_option_is_short: bool = BooleanProperty(defaultvalue=False)
    first_option_background_color = ColorProperty(SUCCESS_COLOR)
    first_option_color = ColorProperty((0, 0, 0, 1))

    second_option_label: str = StringProperty()
    second_option_icon: str = StringProperty()
    second_option_is_short: bool = BooleanProperty(defaultvalue=False)
    second_option_background_color = ColorProperty(DANGER_COLOR)
    second_option_color = ColorProperty((1, 1, 1, 1))

    def get_first_item(self: PromptWidget) -> ActionItem | None:
        """Return the first item of the prompt."""
        if not self.first_option_label:
            return None
        return ActionItem(
            label=self.first_option_label,
            icon=self.first_option_icon,
            action=self.first_option_callback,
            is_short=self.first_option_is_short,
            background_color=self.first_option_background_color,
            color=self.first_option_color,
        )

    def get_second_item(self: PromptWidget) -> ActionItem | None:
        """Return the second item of the prompt."""
        if not self.second_option_label:
            return None
        return ActionItem(
            label=self.second_option_label,
            icon=self.second_option_icon,
            action=self.second_option_callback,
            is_short=self.second_option_is_short,
            background_color=self.second_option_background_color,
            color=self.second_option_color,
        )

    first_item: ActionItem = AliasProperty(
        getter=get_first_item,
        bind=[
            'first_option_label',
            'first_option_icon',
            'first_option_is_short',
            'first_option_background_color',
            'first_option_color',
        ],
        cache=True,
    )

    second_item: ActionItem = AliasProperty(
        getter=get_second_item,
        bind=[
            'second_option_label',
            'second_option_icon',
            'second_option_is_short',
            'second_option_background_color',
            'second_option_color',
        ],
        cache=True,
    )

    def _items_getter(self: PromptWidget) -> Sequence[Item | None]:
        return [self.first_item, self.second_item]

    def _items_setter(self: PromptWidget, value: Sequence[Item | None]) -> None:
        if len(value) == 0:
            return
        msg = 'Cannot set items'
        raise ValueError(msg)

    items: Sequence[Item | None] = AliasProperty(
        getter=_items_getter,
        setter=_items_setter,
        bind=['first_item', 'second_item'],
    )

    @abstractmethod
    def first_option_callback(
        self: PromptWidget,
    ) -> Menu | type[PageWidget] | None:
        """Do what the first option of the prompt should do in this callback.

        It can return a `SubMenuItem` or an `ApplicationItem` and it will become active
        if this prompt is rendered in a `MenuWidget`.
        """
        return

    @abstractmethod
    def second_option_callback(
        self: PromptWidget,
    ) -> Menu | type[PageWidget] | None:
        """Do what the second option of the prompt should do in this callback.

        It can return a `SubMenuItem` or an `ApplicationItem` and it will become active
        if this prompt is rendered in a `MenuWidget`.
        """
        return

    def get_item(self: PromptWidget, index: int) -> Item | None:
        """Return the page item at the given index."""
        if not 1 <= index <= PROMPT_OPTIONS:
            warnings.warn('index must be either 1 or 2', ResourceWarning, stacklevel=1)
            return None
        return self.first_item if index == 1 else self.second_item


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('prompt_widget.kv').resolve().as_posix(),
)



"""Implement `ItemWidget`."""

from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING, Any, Callable

from kivy.lang.builder import Builder
from kivy.properties import (
    BooleanProperty,
    ColorProperty,
    NumericProperty,
    ObjectProperty,
    StringProperty,
)
from kivy.uix.boxlayout import BoxLayout

from ubo_gui.constants import PRIMARY_COLOR
from ubo_gui.menu.types import process_subscribable_value

if TYPE_CHECKING:
    from kivy.graphics.context_instructions import Color

    from ubo_gui.menu.types import Item


class ItemWidget(BoxLayout):
    """Renders an `Item`.

    Attributes
    ----------
    color: `tuple` of `int`
        Color of the item, used for its text and icon.

    background_color: `tuple` of `int`
        Background color of the item.

    label:
        The text rendered in the item.

    icon: `str`
        Name of a Material Symbols icon.

    """

    _subscriptions: list[Callable[[], None]]

    is_set: bool = BooleanProperty(defaultvalue=False)
    label: str = StringProperty()
    color: Color = ColorProperty((1, 1, 1, 1))
    background_color: Color = ColorProperty(PRIMARY_COLOR)
    icon: str = StringProperty(defaultvalue='')
    is_short: bool = BooleanProperty(defaultvalue=False)
    item: Item | None = ObjectProperty(allownone=True)
    opacity: float = NumericProperty(defaultvalue=1, min=0, max=1)
    progress: float = NumericProperty(defaultvalue=1, min=0, max=1)

    _width: float = NumericProperty()
    _progress: int = NumericProperty()

    def __init__(self: ItemWidget, item: Item | None = None, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize an `ItemWidget`."""
        self._subscriptions = []
        super().__init__(item=item, **kwargs)

    def __del__(self: ItemWidget) -> None:
        """Unsubscribe from the item."""
        self.clear_subscriptions()

    def on_item(self: ItemWidget, _: ItemWidget, value: Item | None) -> None:
        """Update the widget properties when the item changes."""
        self.clear_subscriptions()
        if value is None:
            self.is_set = False
        else:
            self.is_set = True
            self.label = ''
            self._subscriptions.append(
                process_subscribable_value(
                    value.label,
                    lambda value: setattr(self, 'label', value or ''),
                ),
            )

            self.is_short = False
            self._subscriptions.append(
                process_subscribable_value(
                    value.is_short,
                    lambda value: setattr(
                        self,
                        'is_short',
                        False if value is None else value,
                    ),
                ),
            )

            self.color = ItemWidget.color.defaultvalue
            self._subscriptions.append(
                process_subscribable_value(
                    value.color,
                    lambda value: setattr(
                        self,
                        'color',
                        value or ItemWidget.color.defaultvalue,
                    ),
                ),
            )

            self.background_color = ItemWidget.background_color.defaultvalue
            self._subscriptions.append(
                process_subscribable_value(
                    value.background_color,
                    lambda value: setattr(
                        self,
                        'background_color',
                        value or ItemWidget.background_color.defaultvalue,
                    ),
                ),
            )

            self.icon = ''
            self._subscriptions.append(
                process_subscribable_value(
                    value.icon,
                    lambda value: setattr(self, 'icon', value or ''),
                ),
            )

            self.opacity = value.opacity or 1
            self.progress = min(max(value.progress or 1, 0), 1)

    def clear_subscriptions(self: ItemWidget) -> None:
        """Clear the subscriptions."""
        for subscription in self._subscriptions:
            subscription()
        self._subscriptions.clear()


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('item_widget.kv').resolve().as_posix(),
)



"""Stack item classes for the menu stack."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Callable, Self, TypeVar, cast

if TYPE_CHECKING:
    from ubo_gui.page import PageWidget

    from .types import Menu

VISUAL_SNAPSHOT_WIDTH = 15


@dataclass(kw_only=True)
class BaseStackItem:
    """An item in a menu stack."""

    parent: StackItem | None
    subscriptions: set[Callable[[], None]] = field(default_factory=set)

    def clear_subscriptions(self: BaseStackItem) -> None:
        """Clear all subscriptions."""
        subscriptions = self.subscriptions.copy()
        self.subscriptions.clear()
        for unsubscribe in subscriptions:
            unsubscribe()

    @property
    def root(self: BaseStackItem) -> StackMenuItem:
        """Return the root item."""
        if self.parent:
            return self.parent.root
        return cast(StackMenuItem, self)

    @property
    def title(self: Self) -> str:
        """Return the title of the item."""
        raise NotImplementedError


@dataclass(kw_only=True)
class StackMenuItemSelection:
    """A selected menu item in a menu stack."""

    key: str
    item: StackMenuItem


@dataclass(kw_only=True)
class StackMenuItem(BaseStackItem):
    """A menu item in a menu stack."""

    menu: Menu
    page_index: int
    selection: StackMenuItemSelection | None = None

    @property
    def title(self: StackMenuItem) -> str:
        """Return the title of the menu."""
        return self.menu.title() if callable(self.menu.title) else self.menu.title

    @property
    def visual_snapshot(self: StackMenuItem) -> list[str]:
        """Return the snapshot of the menu."""
        T = TypeVar('T')

        def process_callable(object_: T | Callable[[], T]) -> T:
            return object_() if callable(object_) else object_

        items = process_callable(self.menu.items)
        title = process_callable(self.menu.title)[: VISUAL_SNAPSHOT_WIDTH - 2]
        padding = '─' * ((VISUAL_SNAPSHOT_WIDTH - len(title)) // 2)
        return [
            padding
            + title
            + padding
            + '─' * ((len(title) + VISUAL_SNAPSHOT_WIDTH) % 2),
            *(
                (
                    (process_callable(items[i].icon) or ' ')
                    + ' '
                    + str(process_callable(items[i].label))
                ).ljust(
                    VISUAL_SNAPSHOT_WIDTH,
                    ' ',
                )[:VISUAL_SNAPSHOT_WIDTH]
                if len(items) > i
                else ' ' * VISUAL_SNAPSHOT_WIDTH
                for i in range(3)
            ),
            '─' * VISUAL_SNAPSHOT_WIDTH,
        ]


@dataclass(kw_only=True)
class StackApplicationItem(BaseStackItem):
    """An application item in a menu stack."""

    application: PageWidget

    @property
    def title(self: StackApplicationItem) -> str:
        """Return the title of the application."""
        return self.application.name

    @property
    def visual_snapshot(self: StackApplicationItem) -> list[str]:
        """Return the snapshot of the application."""
        title = self.title[: VISUAL_SNAPSHOT_WIDTH - 2]
        padding = '─' * ((VISUAL_SNAPSHOT_WIDTH - len(title)) // 2)
        return [
            padding + title + padding + '─' * (len(title) % 2),
            ' ' * VISUAL_SNAPSHOT_WIDTH,
            f""" {(self.application.title or '-').ljust(VISUAL_SNAPSHOT_WIDTH - 2, " ")
            [:VISUAL_SNAPSHOT_WIDTH - 2]} """,
            ' ' * VISUAL_SNAPSHOT_WIDTH,
            '─' * VISUAL_SNAPSHOT_WIDTH,
        ]


StackItem = StackMenuItem | StackApplicationItem



"""A half circle gauge widget."""

from __future__ import annotations

import pathlib

from kivy.lang.builder import Builder
from kivy.properties import ColorProperty, ListProperty, NumericProperty, StringProperty
from kivy.uix.boxlayout import BoxLayout


class GaugeWidget(BoxLayout):
    """A widget that displays a gauge."""

    value = NumericProperty(50)
    min_value = NumericProperty(0)
    max_value = NumericProperty(100)
    fill_color = ColorProperty('red')
    background_color = ColorProperty('#D9D9D9')
    label = StringProperty('')

    _size = ListProperty([0, 0])
    _pos = ListProperty([0, 0])


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('gauge_widget.kv').resolve().as_posix(),
)



"""Provides easy access to different transitions."""

from __future__ import annotations

import threading
from functools import cached_property
from typing import Any, NotRequired, TypedDict

from headless_kivy import HeadlessWidget
from kivy.clock import mainthread
from kivy.uix.screenmanager import (
    NoTransition,
    RiseInTransition,
    Screen,
    ScreenManager,
    SlideTransition,
    SwapTransition,
    TransitionBase,
)
from kivy.uix.widget import Widget


class SwitchParameters(TypedDict):
    """Parameters for switching screens."""

    duration: NotRequired[float | None]
    direction: NotRequired[str | None]


class TransitionsMixin:
    """Provides easy access to different transitions."""

    transition_queue: list[
        tuple[Screen | None, TransitionBase, str | None, float | None]
    ]
    screen_manager: ScreenManager
    _is_transition_in_progress: bool = False
    _is_preparation_in_progress: bool = False
    _transition_progress_lock: threading.Lock

    def __init__(self: TransitionsMixin, **kwargs: dict[str, Any]) -> None:
        """Initialize the transitions mixin."""
        _ = kwargs
        self._transition_progress_lock = threading.Lock()
        self.transition_queue = []

    def _handle_transition_progress(
        self: TransitionsMixin,
        transition: TransitionBase,
        progression: float,
    ) -> None:
        transition.screen_out.opacity = 1 - progression
        transition.screen_in.opacity = progression

    def _handle_transition_complete(
        self: TransitionsMixin,
        transition: TransitionBase,
    ) -> None:
        transition.screen_out.opacity = 0
        transition.screen_in.opacity = 1
        with self._transition_progress_lock:
            if self.transition_queue:
                (
                    (screen, transition, direction, duration),
                    *self.transition_queue,
                ) = self.transition_queue
                if (
                    len(self.transition_queue) > 1
                    and transition is not self._no_transition
                ):
                    duration = 0.08
                switch_parameters: SwitchParameters = {}
                if duration is not None:
                    switch_parameters['duration'] = duration
                if direction is not None:
                    switch_parameters['direction'] = direction
                self._perform_switch(
                    screen,
                    transition=transition,
                    **switch_parameters,
                )
            else:
                if isinstance(self, Widget):
                    headless_widget = HeadlessWidget.get_instance(self)
                    if headless_widget:
                        headless_widget.activate_low_fps_mode()
                self._is_transition_in_progress = False

    def _setup_transition(self: TransitionsMixin, transition: TransitionBase) -> None:
        transition.bind(on_progress=self._handle_transition_progress)
        transition.bind(on_complete=self._handle_transition_complete)

    @cached_property
    def _no_transition(self: TransitionsMixin) -> NoTransition:
        transition = NoTransition()
        self._setup_transition(transition)
        return transition

    @cached_property
    def _slide_transition(self: TransitionsMixin) -> SlideTransition:
        transition = SlideTransition()
        self._setup_transition(transition)
        return transition

    @cached_property
    def _rise_in_transition(self: TransitionsMixin) -> RiseInTransition:
        transition = RiseInTransition()
        self._setup_transition(transition)
        return transition

    @cached_property
    def _swap_transition(self: TransitionsMixin) -> SwapTransition:
        transition = SwapTransition()
        self._setup_transition(transition)
        return transition

    @mainthread
    def _perform_switch(
        self: TransitionsMixin,
        screen: Screen | None,
        /,
        *,
        transition: TransitionBase,
        duration: float | None = None,
        direction: str | None = None,
    ) -> None:
        if duration is None:
            duration = 0.2
        self.screen_manager.switch_to(
            screen,
            transition=transition,
            duration=duration,
            **({} if direction is None else {'direction': direction}),
        )
        self._is_preparation_in_progress = False

    def _switch_to(
        self: TransitionsMixin,
        screen: Screen | None,
        /,
        *,
        transition: TransitionBase,
        duration: float | None = None,
        direction: str | None = None,
    ) -> None:
        """Switch to a new screen."""
        if screen is self.screen_manager.current_screen:
            return
        if duration is None:
            duration = 0 if transition is self._no_transition else 0.3
        with self._transition_progress_lock:
            if self._is_transition_in_progress:
                self.transition_queue = [
                    *self.transition_queue,
                    (screen, transition, direction, duration),
                ]
            else:
                if isinstance(self, Widget):
                    headless_widget = HeadlessWidget.get_instance(self)
                    if headless_widget:
                        headless_widget.activate_high_fps_mode()
                self._is_transition_in_progress = transition is not self._no_transition
                self._is_preparation_in_progress = True
                self._perform_switch(
                    screen,
                    transition=transition,
                    duration=duration,
                    direction=direction,
                )
